//------------------------------------------------------------------------
//
// Copyright (C) 1997-2006 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module:  C++ Example
// 
// Description
//
// This is a sample C++ application which demonstrates using the Chameleon
// API to receive and parse HL7 messages over an LLP connection.
//
// Author: iNTERFACEWARE
// Date:   Wed 09/09/1998 
//---------------------------------------------------------------------------
#include "CHMserverPreCompHead.h"
#pragma hdrstop

#include <cstdlib>

#include <iostream>

#include <CHM/CHMllpServer.h>
#include <CHM/CHMerrorClass.h>
#include <CHM/CHMllpConnection.h>
#include <CHM/CHMstring.h>

using namespace std;

/// This application is a simple example of using the Chameleon generated objects.  It
/// is intended to be a helpful starting point in understanding how to use these objects
/// but must be fleshed out in the following areas to provide a full solution for an
/// application HL7 interface:
/// 1) You will need to work out how your application will handle logging of messages
///    and data, in particular how bad transactions should be logged for effective troubleshooting.
/// 2) The code will need to be modified to insert the parsed data into a database whether
///    by using explicit database calls or going through some type of middle tier layer that
///    your application may support.
/// 3) Some customers choose to deploy HL7 interfaces as Windows Services.  This can be done
///    easily with Chameleon.  One thing to keep in mind though is that the networking components
///    supplied with Chameleon require a Windows 'message loop'.  Often Windows Services do
///    not have a message loop - but one can be easily set up by invoking RunMessageLoop on
///    the networking component supplied with Chameleon.  See the manual for running Chameleon
///    within services.
/// 4) Definitely take the time to read the manual and go through the tutorials after looking
///    at this example before attempting to deploy a production solution.  We do provide custom
///    on-site training and technical implementation services if you require them.
///
/// If you are looking for a much more complete 'out of the box' solution for your HL7 needs it would be
/// a good idea to evaluate iNTERFACEWARE Iguana.  Iguana is powered by Chameleon but
/// provides a complete run-time web-based infrastructure for your HL7 interfaces.
/// It provides advanced logging and monitoring capability which makes it a very attractive option 
/// for many customers.

#define NEWLINE "\r\n"
#define MAX_MESSAGE_LENGTH 200
#define GETLINE_BUFFER_SIZE 100

#define MAX_CONNECTIONS 16
#define DEFAULT_PORT 5123

class CHMserver : public CHMllpServer
{
public:

   CoutOutputLog Log;
   CHMengine AutoAckEngine;  // engine instance we use for automatic ACKs (dynamic)
   CHMengine AutoNackEngine; // engine instance we use for negative acknowledgements (dynamic)
   CHMstring Flatwire;
   PrefixTestappEngine Engine;

   // This is the event we will get when a Client is connected to the Server.
   // The 'Client' object (of type CHMllpConnection) can be used to interact with the machine that is sending
   // the data.  There can be multiple clients connected to the server.
   virtual void OnClientConnected(CHMllpConnection& Client)
   {
      CHMstring Output = "Accepted connection from ";
      Output += Client.RemoteIpAddressAsString();
      Output += "\nCurrent number of client connections: ";
      CHMstring NumConnection(MAX_CONNECTIONS, 0);
      sprintf(NumConnection.get_buffer(), "%d", this->CountOfClient());
     
      Output += NumConnection;
    
      Log.OutputLine(MakeMessagePrintable(Output.c_str(), NEWLINE, MAX_MESSAGE_LENGTH));
      Log.OutputLine("");     
   }

   // This is the event we will get when a Client disconnects from the server. Note that sometimes
   // if the client is not well behaved we may not receive this event.
   virtual void OnClientDisconnected(CHMllpConnection& Client )
   {
      CHMstring Output = "Disconnected from ";
      Output += Client.RemoteIpAddressAsString();
     
      Log.OutputLine(MakeMessagePrintable(Output.c_str(), NEWLINE, MAX_MESSAGE_LENGTH));
      Log.OutputLine("");
   }

   // This event is the main entry point when we get an HL7 message from a connected 
   // peer.  We invoke the ParseMessage method on the hl7Engine object.  
   //
   // Note that the LLP TCP/IP transport components are not tightly coupled to the
   // hl7Engine object.  You can use any kind of transport you like with Chameleon, including
   // your own transports or third party components.
   //
   // Notice that in this application we do not block on events - this is because the
   // networking components are asynchronous.  You can write synchronous applications using
   // Chameleon but you would need to use a transport mechanism other than the LLP TCP/IP
   // components supplied here.
   virtual void OnMessageReceived(CHMllpConnection& Client,const char* Message)
   {
      Log.OutputLine("Received a message: ");
      try
      {
         // We manually invoke the engine to parse the message.  The second
         // parameter is set to 'null'; it used to be part of the transport framework for
         // Chameleon, but is no longer required.
         Engine.ParseMessage(Message, 0);
      }
      catch(CHMerror& Error)
      {
         Log.OutputLine("An error occurred while parsing the message: ");
         Log.OutputLine(Error.Description()); 
      }
      catch (...)
      {
         Log.OutputLine("An unhandled exception occurred while parsing the message ");
      }
      SendAckToClient(Client, Message);
   }

   // This is the event we'll get when we have an error on a given TCP/IP connection.
   virtual void OnClientErrorReceived(CHMllpConnection& Client, const char* ErrorMessage, CHMint32 ErrorCode)
   {
      CHMstring Output = "A connection error occurred with client ";
      Output += Client.RemoteIpAddressAsString();
      Output += " : ";
      Output += ErrorMessage;
      Log.OutputLine(Output.c_str());

   }

   // This event fires if we have an error related to the listening socket itself rather
   // than a specific connection.  In practice this event is rarely invoked.
   virtual void OnErrorReceived(const char* ErrorMessage, CHMint32 Code)
   {
      Log.OutputLine("A socket error has occurred.");
      // Disconnect all clients on server error
      DisconnectAllClients();
      Log.OutputLine(ErrorMessage);
   }

   // This event fires when we receive data that is not using the Lower Layer Protocol (LLP).
   // LLP is the standard protocol for sending HL7 messages, and is used by Chameleon.  See the manual
   // for more details.
   virtual void OnDataIgnored(CHMllpConnection& Client, const char* Data)
   {
      CHMstring Output = "Received ignored data ";
      Output += Data;
      Log.OutputLine(Output.c_str());
   }

   void SendAckToClient(CHMllpConnection& Client, const char* Message)
   {
      const char* Ack;
      try
      {
         Log.OutputLine("");
         // We are using auto-acknowledgement - see the manual for information  
         // Now we need to generate the ack and send it back to the user.
         // We use auto-acknowledgement; to obtain more information regarding auto-acknowledgement,
         // please read the Chameleon manual.

         //this method generates an Ack using the ackEngine
         Ack = AutoAckEngine.generateAck(Message);      
         Log.OutputLine("ACK generated: "); 
         // We use the MakeMessagePrintable function to 'pretty print' the HL7 message 
         // into a format that can be easily read in the log files.  This function converts 
         // carriage return (\r) characters into the string given in the second argument of 
         // MakeMessagePrintable.  In this case it is converted into a carriage return/line feed
         // (\r\n) sequence which displays nicely in the memo control.
         // 
         // The third argument is the maximum length of the message before
         // we truncate it.  This is done because displaying data within a memo
         // control is actually very slow - so displaying a message of 100k in
         // the memo control window can give the appearance of hanging the server
         // (i.e. a long blocking event).
         //
         // Having the third argument set to -1 means we are not truncating the output. 
         Log.OutputLine(MakeMessagePrintable(Ack, NEWLINE, MAX_MESSAGE_LENGTH));
         Log.OutputLine("Sending ACK back to client.");

      }
      catch (CHMerror& Error)
      {
         Log.OutputLine("An error occured while generating an ACK.");
         Log.OutputLine(Error.Description());
         Log.OutputLine("Sending a NACK instead ");
         try
         {
            // An error occured! Send back a negative-acknowledgement
            Ack = AutoNackEngine.generateAck(Message);
         }
         catch (CHMerror& Error)
         {
            // Just in case generating the Nack fails, we'll send a hard-coded nack instead.
            Log.OutputLine("NACK generation failed. Sending a hard-coded NACK.");
            Ack = "MSH|^~\\&|ACME|ACME_APP|XYZ|XYZ_APP|||NACK|XXX|P|2.4|\rMSA|AA|XXX|\r";
         }
      }
      catch (...)
      {
         Log.OutputLine("Unhandled exception during ACK generation.");
      }
      
      Client.TransmitMessage(Ack);
   }

};

int main()
{  
   try
   {
   
      CHMserver Listener;
      // In order to make the engine usable we must load the vmd file.  This file
      // contains the information about the structure of the messages, tables, and data
      // mappings that were defined in the Chameleon GUI.  Listener.Engine is a generated 
      // Engine which has an interface based on the messages and tables defined in the vmd file.
      // 
      // Also keep in mind that this example is using the static stub-code 
      // interface of Chameleon. For advanced applications it is possible to 
      // use a dynamic interface to populate messages.  See the manual for 
      // more details.      
      Listener.Engine.Load("test_data.vmd");
      Listener.Engine.SetLog(new CoutOutputLog());

      cout << "Loaded message definition file test_data.vmd" << endl;

      // We also initialize the engine used for auto-acknowledgment/negative acknowledgment.
      // We use 2 more engines to generate HL7 acknowledgment (ACK) and negative 
      // acknowledgment (NACK) messages. Here the Engines we use are DynamicEngine 
      Listener.AutoAckEngine.Load("autoack.vmd"); 
      Listener.AutoNackEngine.Load("autonack.vmd"); 
      cout << "Loaded automatic ACK message definition file" << endl;
      cout << "*******************************************" << endl;
      cout << "Please select a configuration (default is 0) " << endl;
      for (int ConfigIndex = 0; ConfigIndex < Listener.Engine.CountOfConfiguration(); ConfigIndex++) 
      {
          cout << "# " << ConfigIndex << ". " << Listener.Engine.ConfigurationName(ConfigIndex) << endl;
      }
      Listener.Engine.SetCurrentConfigurationIndex(0);
      char Input[GETLINE_BUFFER_SIZE];
      int Configuration =0;
      cin.getline(Input, GETLINE_BUFFER_SIZE);
      if(Input[0] != 0)
      {      
         Configuration = atoi(Input);
      }
      if ( Configuration > Listener.Engine.CountOfConfiguration() || Configuration < 0)
      {
         cout << "An invalid configuration was entered. Using default configuration (0)." << endl;
      }
      else
      {
          cout << "Current configuration changed to " << Configuration << endl;
          Listener.Engine.SetCurrentConfigurationIndex(Configuration);
      }
      cout << "Enter the port number to listen on (default is " << DEFAULT_PORT << ")  " << endl;
      int Port = DEFAULT_PORT;
      cin.getline(Input, GETLINE_BUFFER_SIZE);
      if(Input[0] != 0)
      {
         Port = atoi(Input);
      }
      if(Port == 0)
      {
         cout << "An invalid port was entered. Listening on default port " << DEFAULT_PORT << "." << endl;
         Port = DEFAULT_PORT;
      }
     
      cout << "Attempting to listen on port " << Port << endl;
      Listener.StartListen(Port);


      cout << "Listening on port " << Listener.Port() << endl << endl;
      // Calling runMessageLoop to start the listener
      // NOTE: this is a blocking call, it will be stopped by entering 'ctrl + C'
      cout << "Press 'Ctrl + C' to terminate the server" << endl; 
      CHMllpServer::RunMessageLoop();
      
      // We received a 'Ctrl+C' event, stop listening
      Listener.StopListen();
   }
   catch (CHMerror& Error)
   {
      printf(Error.Description());
      return EXIT_FAILURE;
   }
   return EXIT_SUCCESS;
}
